@@ -132,16 +132,22 @@ class Byneuron:
     @property
     def token(self):
         """triggers login if needed"""
-        if self.now > self._token_expire:
+        n = self.now
+        if not self._token or n > self._token_expire:
             data = self._login()
             if data:
                 log.debug('set _token from data \n%s ', json.dumps(data, indent=2))
                 self._token = data.get('access_token', '')
-                self._token_expire = self.now + datetime.timedelta(seconds=data.get('expires_in', 0))
+                self._token_expire = n + datetime.timedelta(seconds=data.get('expires_in', 0))
         return self._token
 
+    def del_token(self):
+        """removes token, will force new login on next url"""
+        self._token = ''
+        self._token_expire = datetime.datetime.min.replace(tzinfo=datetime.timezone.utc)
+
     def url(self, url, headers=None, data=None, as_json=True):
-        log.info('requests url %s', url)
+        log.debug('requests url %s', url)
         if headers is None:
             headers = self.headers
         if isinstance(data, dict):
@@ -284,6 +290,11 @@ class Byneuron:
     def _data_sort(self, order):
         return 'DESC' if order == 'DESC' else 'ASC'
     
+    def _data_timestamp(self, dt):
+        if isinstance(dt,datetime.datetime):
+            return int(1000 * dt.timestamp())
+        
+        
     def _numberevents(self, size=100, order=None, deleted=False, indexsets=None, filters=None):
         """
         :param size:
@@ -340,8 +351,8 @@ class Byneuron:
         })
         filters.append({
             "type": "range",  # this is sended as be UTC
-            "from": int(1000 * from_date.timestamp()),
-            "to": int(1000 * to_date.timestamp()),
+            "from": self._data_timestamp(from_date),
+            "to": self._data_timestamp(to_date),
             "minInclusive": True,
             "maxInclusive": False,
             "invertFilter": False
@@ -352,9 +363,9 @@ class Byneuron:
             "time_zone": "UTC",  # "Africa/Nairobi" >> important on how we wrap results
             "interval": interval,
             "name": "datehistogram",
-            "from": int(1000 * from_date.timestamp()), 
+            "from": self._data_timestamp(from_date),
             # wierldly, this 'from' is a requirement to retrieve data with gapstrategy on
-            "to": int(1000 * to_date.timestamp()),
+            "to": self._data_timestamp(to_date),
             "aggregations": aggrs
         }
         data = {
@@ -376,20 +387,40 @@ class Byneuron:
         return self.url(url=url, data=data)
 
     ### endpoint handlers ###
-    def get_numberevents(self, item, size):
-        """ last 100 events for item in {datetime:value} format """
+    def get_numberevents(self, item, size, date=None, order='DESC'):
+        """
+        last 100 (size) events for item in {datetime:value} format
+        starts at current timestamp (or date) and returns in time (or goes in future with order 'ASC')
+
+        """
+
         items = item if isinstance(item, list) else [item]
         item_keys = list({i.key if isinstance(i, Entity) else i for i in items})
         indexset_keys = list({i.indexset if isinstance(i, Entity) else self.indexset_active.key for i in items})
         log.debug('get numberevents for in indexset %s items %s ', indexset_keys, item_keys)
-        f = {
+        f_item = {
             "type": "eventitemid",
             "values": item_keys,
             "invertFilter": False,
             "operator": "or"
         }
+        dt = date if isinstance(date, datetime.datetime) else self.now
+        order = self._data_sort(order)
+        ts = self._data_timestamp(dt)
+        if order == 'DESC':
+            f_time = {  # added to return only historical events.
+                "type": "range",
+                "to": ts,
+                "maxInclusive": True
+            }
+        elif order == 'ASC':
+            f_time = {  # added to return only future events.
+                "type": "range",
+                "from": ts,
+                "minInclusive": True
+            }
         indexsets = indexset_keys
-        data = self._numberevents(indexsets=indexsets, filters=[f], size=size)
+        data = self._numberevents(indexsets=indexsets, filters=[f_item, f_time], size=size, order='DESC')
         if data:
             for i in data:
                 yield Numberevent(datamodel=i)
@@ -411,7 +442,10 @@ class Byneuron:
             from_date = self.now
         if not isinstance(to_date, datetime.datetime):
             to_date = self.now
-        order = self._data_sort(order)
+        if order == 'ASC':
+            log.warning('ASC sorting not allowed for get_numberevents_dates !')
+
+        order = 'DESC' # self._data_sort(order)  # attention, must be 'DESC'
         items = item if isinstance(item, list) else [item]
         item_keys = list({i.key if isinstance(i, Entity) else i for i in items})
         indexset_keys = self._data_indexsets(items)
@@ -423,12 +457,12 @@ class Byneuron:
         }
         f_range = {
             "type": "range",
-            "to": int(max(from_date, to_date).timestamp() * 1000),
+            "to": self._data_timestamp(max(from_date, to_date)),
             "maxInclusive": False
         }
         if not extend:
             f_range.update({
-                "from": int(min(from_date, to_date).timestamp() * 1000),
+                "from": self._data_timestamp(min(from_date, to_date)),
                 "minInclusive": True
             })
         while True:
@@ -440,7 +474,7 @@ class Byneuron:
                     if extend and event.datetime < min(from_date, to_date):
                         return
                 f_range.update({
-                    "to": event.datetime.timestamp() * 1000,  # this is why we must have sorting to DESC
+                    "to": self._data_timestamp(event.datetime),  # this is why we must have sorting to DESC
                     "maxInclusive": False
                 })
             else:
@@ -636,6 +670,11 @@ class Byneuron:
         return [e for e in r.values() if isinstance(e, Entity)]
 
     def datamodel(self, data):
+        """
+        attention this is an iterator
+        :param data: (list with ) Numberevent, Entity
+        :return: yields the datamodel objects
+        """
         data = self._graphmodel(data)
         if data:
             log.debug('graphmodel returned %s', data)
@@ -822,7 +861,7 @@ class Byneuron:
             self._isc[codetype].append(code)
             # create code
 
-        log.info('buffered indexSetCodetypes %s', self._isc)
+        log.debug('buffered indexSetCodetypes %s', self._isc)
         # buffer existing codes
 
 
@@ -1169,7 +1208,8 @@ class Numberevent:
             # if not timezone aware; force UTC
             if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
                 dt.replace(tzinfo=datetime.timezone.utc)
-            return dt.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
+            dt_utc = dt.astimezone(datetime.timezone.utc)
+            return dt_utc.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'  # microseconds truncated to milliseconds
 
         datamodel = {
             "timestamp": timestamp(),
